/*
** pork_events.c - event handler functions.
** Copyright (C) 2002-2003 Amber Adams <amber@ojnk.net>
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <sys/types.h>
#include <ncurses.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include <pork.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_set.h>
#include <pork_screen_cmd.h>
#include <pork_perl.h>
#include <pork_events.h>
#include <pork_command.h>
#include <pork_command_defs.h>

static const struct event_info events_info[] = {
	{ "BUDDY_AWAY",				"SS"		},
	{ "BUDDY_BACK",				"SS"		},
	{ "BUDDY_IDLE",				"SiS"		},
	{ "BUDDY_SIGNOFF",			"SS"		},
	{ "BUDDY_SIGNON",			"SS"		},
	{ "BUDDY_UNIDLE",			"SS"		},
	{ "QUIT",					""			},
	{ "RECV_AWAYMSG",			"SiiiisS"	},
	{ "RECV_CHAT_INVITE",		"SSisS"		},
	{ "RECV_CHAT_JOIN",			"SSiSS"		},
	{ "RECV_CHAT_LEAVE",		"SSiSS"		},
	{ "RECV_CHAT_MSG",			"SSiSsS"	},
	{ "RECV_IM",				"SsS"		},
	{ "RECV_PROFILE",			"SiiisS"	},
	{ "RECV_SEARCH_RESULT",		"SSS"		},
	{ "RECV_WARN",				"SiS"		},
	{ "SEND_AWAY",				"sS"		},
	{ "SEND_CHAT_INVITE",		"SiSsS"		},
	{ "SEND_CHAT_JOIN",			"SiS"		},
	{ "SEND_CHAT_LEAVE",		"SSiS"		},
	{ "SEND_CHAT_MSG",			"SSisS"		},
	{ "SEND_IDLE",				"iS"		},
	{ "SEND_IM",				"SsS"		},
	{ "SEND_LINE",				"sS"		},
	{ "SEND_PROFILE",			"sS"		},
	{ "SEND_WARN",				"SiS"		},
	{ "SIGNOFF",				"S"			},
	{ "SIGNON",					"S"			},
};

static u_int32_t refnum;

static int execute_pork_cmd(char *cmd, const char *fmt, va_list ap);

static int events_compare(const void *l, const void *r) {
	const char *key = (char *) l;
	struct event_info *info = (struct event_info *) r;

	return (strcasecmp(key, info->name));
}

static int events_find_refnum_cb(void *l, void *r) {
	u_int32_t refnum = (u_int32_t) l;
	struct event_entry *event = (struct event_entry *) r;

	return (refnum - event->refnum);
}

static int event_find(const char *name) {
	struct event_info *info;
	u_int32_t offset;

	info = bsearch(name, events_info, array_elem(events_info),
			sizeof(struct event_info), events_compare);

	if (info == NULL)
		return (-1);

	offset = (u_int32_t) info - (u_int32_t) &events_info[0];
	return (offset / sizeof(struct event_info));
}

inline void events_init(struct event *events) {
	memset(events, 0, sizeof(*events));
}

static void event_destroy_cb(void *param __notused, void *data) {
	struct event_entry *event = (struct event_entry *) data;

	free(event->command);
}

inline void events_purge(struct event *events) {
	events_destroy(events);
	events_init(events);
}

void events_destroy(struct event *events) {
	u_int32_t i;

	for (i = 0 ; i < MAX_EVENT_TYPE ; i++) {
		dlist_destroy(events->e[i], NULL, event_destroy_cb);
		events->e[i] = NULL;
	}
}

void list_events(struct event *events, const char *event_type) {
	u_int32_t e;

	if (event_type != NULL) {
		int event;

		event = event_find(event_type);
		if (event == -1) {
			screen_err_msg("No such event: %s", event_type);
			return;
		}

		print_event_type(events, event);
		return;
	}

	for (e = 0 ; e < MAX_EVENT_TYPE ; e++)
		print_event_type(events, e);
}

void print_event_type(struct event *events, u_int32_t e) {
	dlist_t *cur = events->e[e];

	if (cur == NULL) {
		screen_output("no %s events", events_info[e].name);
		return;
	}

	do {
		struct event_entry *event = cur->data;

		screen_output("%u ON %s: %s",
			event->refnum,
			events_info[e].name,
 			event->command);

		cur = cur->next;
	} while (cur != NULL);
}

int add_event(struct event *events, const char *event_type, char *args) {
	int event;
	int quiet = 0;
	struct event_entry *new_event;

	if (*event_type == '^') {
		quiet = 1;
		event_type++;
	}

	event = event_find(event_type);
	if (event == -1)
		return (-1);

	new_event = xcalloc(1, sizeof(*new_event));
	new_event->command = xstrdup(args);
	new_event->quiet = quiet;
	new_event->refnum = refnum++;

	events->e[event] = dlist_add_tail(events->e[event], new_event);
	return (0);
}

int events_del_refnum(struct event *events, u_int32_t refnum) {
	dlist_t *cur;
	u_int32_t i;

	for (i = 0 ; i < MAX_EVENT_TYPE ; i++) {
		cur = dlist_find(events->e[i], (void *) refnum, events_find_refnum_cb);
		if (cur != NULL) {
			struct event_entry *event = cur->data;

			events->e[i] = dlist_remove(events->e[i], cur);
			free(event->command);
			free(event);

			return (0);
		} else
			cur = cur->next;
	}

	return (-1);
}

int events_del_type(struct event *events, const char *type, const char *func) {
	int event = event_find(type);
	dlist_t *cur;

	if (event == -1)
		return (-1);

	if (func == NULL) {
		dlist_destroy(events->e[event], NULL, event_destroy_cb);
		events->e[event] = NULL;

		return (0);
	}

	cur = events->e[event];
	while (cur != NULL) {
		struct event_entry *entry = cur->data;

		if (!strcasecmp(entry->command, func)) {
			events->e[event] = dlist_remove(events->e[event], cur);
			free(entry->command);
			free(entry);

			return (0);
		}
	}

	return (-1);
}

int do_event(struct event *events, u_int32_t e, ...) {
	const char *format;
	dlist_t *cur;
	bool quiet = 0;

	format = events_info[e].fmt;

	cur = events->e[e];
	while (cur != NULL) {
		int ret;
		va_list data;
		struct event_entry *event = cur->data;

		if (event->quiet)
			quiet = screen_set_quiet(1);

		va_start(data, e);
		if (event->command[0] == opt_get_char(OPT_CMDCHARS))
			ret = execute_pork_cmd(event->command + 1, format, data);
		else
			ret = execute_perl_va(event->command, format, data);
		va_end(data);

		if (event->quiet)
			screen_set_quiet(quiet);

		if (ret != 0)
			return (ret);

		cur = cur->next;
	}

	return (0);
}

static int execute_pork_cmd(char *cmd, const char *fmt, va_list ap) {
	u_int32_t i;
	u_int32_t s = 0;
	u_int32_t numargs = strlen(fmt);
	char buf[4096];
	size_t buflen = sizeof(buf) - 1;
	char *args[numargs];

	for (i = 0 ; i < numargs ; i++) {
		if (fmt[i] == 'i') {
			u_int32_t i_arg = va_arg(ap, u_int32_t);

			snprintf(buf, sizeof(buf), "%u", i_arg);
			args[i] = xstrdup(buf);
		} else {
			char *val = va_arg(ap, char *);

			if (val == NULL)
				val = "";

			args[i] = xstrdup(val);
		}
	}

	for (i = 0 ; cmd[i] != '\0' && s < buflen ; i++) {
		if (cmd[i] == SYMBOL_VARIABLE) {

			if (isdigit(cmd[++i])) {
				u_int32_t selarg = cmd[i] - '0';

				if (selarg < numargs) {
					u_int32_t j;

					for (j = 0 ; args[selarg][j] != '\0' && s < buflen ; j++)
						buf[s++] = args[selarg][j];
				}

			} else {
				buf[s++] = cmd[--i];
			}

		} else {
			buf[s++] = cmd[i];
		}
	}
	buf[s] = '\0';

	if (*buf != '\0') {
		char *final_command = buf;

		while (*final_command == opt_get_char(OPT_CMDCHARS))
			final_command++;

		run_mcommand(final_command);
	}

	for (i = 0 ; i < numargs ; i++) {
		free(args[i]);
	}

	return (0);
}
