/*
** pork_io.c - I/O handler.
** Copyright (C) 2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <ncurses.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork_util.h>
#include <pork_list.h>
#include <pork_acct.h>
#include <pork_aim.h>
#include <pork_io.h>

static hash_t io_hash;

static int pork_io_find_cb(void *l, void *r) {
	struct io_source *io = (struct io_source *) r;

	return ((u_int32_t) l - (u_int32_t) io->key);
}

static void pork_io_destroy_cb(void *param __notused, void *data) {
	struct io_source *io = (struct io_source *) data;

	/*
	** XXX: may want to do some more things here.
	*/

	free(io);
}

static inline u_int32_t pork_io_hash(u_int32_t key, u_int32_t order) {
	return (key & ((1 << order) - 1));
}

int pork_io_init(void) {
	memset(&io_hash, 0, sizeof(io_hash));

	return (hash_init(&io_hash, 3, pork_io_find_cb, pork_io_destroy_cb));
}

void pork_io_destroy(void) {
	hash_destroy(&io_hash);
}

int pork_io_add(	int fd,
					u_int32_t cond,
					void *data,
					void *key,
					void (*callback)(int fd, u_int32_t cond, void *data))
{
	struct io_source *io;
	u_int32_t hash;

	if (key == NULL)
		return (-1);

	/*
	** If there's already an entry for this key, delete it
	** and replace it with the new one.
	*/

	pork_io_del(key);

	io = xcalloc(1, sizeof(*io));
	io->fd = fd;
	io->cond = cond;
	io->data = data;
	io->key = key;
	io->callback = callback;

	hash = pork_io_hash((u_int32_t) key, io_hash.order);
	hash_add(&io_hash, io, hash);
	return (0);
}

int pork_io_del(void *key) {
	u_int32_t hash = pork_io_hash((u_int32_t) key, io_hash.order);

	return (hash_remove(&io_hash, key, hash));
}

int pork_io_dead(void *key) {
	u_int32_t hash = pork_io_hash((u_int32_t) key, io_hash.order);
	dlist_t *node;

	node = hash_find(&io_hash, key, hash);
	if (node == NULL)
		return (-1);

	((struct io_source *) node->data)->fd = -1;
	return (0);
}

int pork_io_set_cond(void *key, u_int32_t new_cond) {
	u_int32_t hash = pork_io_hash((u_int32_t) key, io_hash.order);
	dlist_t *node;

	node = hash_find(&io_hash, key, hash);
	if (node == NULL)
		return (-1);

	((struct io_source *) node->data)->cond = new_cond;
	return (0);
}

int pork_io_run(void) {
	fd_set rfds;
	fd_set wfds;
	fd_set xfds;
	u_int32_t i;
	int max_fd = -1;
	int ret;
	struct timeval tv = { 0, 500000 };

	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	FD_ZERO(&xfds);

	for (i = 0 ; i < (u_int32_t) (1 << io_hash.order) ; i++) {
		dlist_t *cur = io_hash.map[i];

		while (cur != NULL) {
			struct io_source *io = cur->data;
			dlist_t *next = cur->next;

			if (io->fd >= 0) {
				if (io->cond & IO_COND_READ)
					FD_SET(io->fd, &rfds);

				if (io->cond & IO_COND_WRITE)
					FD_SET(io->fd, &wfds);

				if (io->cond & IO_COND_EXCEPTION)
					FD_SET(io->fd, &xfds);

				if (io->fd > max_fd)
					max_fd = io->fd;
			} else {
				/*
				** If there's a callback, it's responsible for
				** deleting the io source from the set.
				*/

				if (io->callback != NULL)
					io->callback(io->fd, IO_COND_DEAD, io->data);
				else
					pork_io_del(io->key);
			}

			cur = next;
		}
	}

	ret = select(max_fd + 1, &rfds, &wfds, &xfds, &tv);
	if (ret < 1)
		return (ret);

	for (i = 0 ; i < (u_int32_t) (1 << io_hash.order) ; i++) {
		dlist_t *cur = io_hash.map[i];

		while (cur != NULL) {
			struct io_source *io = cur->data;
			dlist_t *next = cur->next;

			if (io->fd >= 0) {
				u_int32_t cond = 0;

				if ((io->cond & IO_COND_READ) && FD_ISSET(io->fd, &rfds))
					cond |= IO_COND_READ;

				if ((io->cond & IO_COND_WRITE) && FD_ISSET(io->fd, &wfds))
					cond |= IO_COND_WRITE;

				if ((io->cond & IO_COND_EXCEPTION) && FD_ISSET(io->fd, &xfds))
					cond |= IO_COND_EXCEPTION;

				if (cond != 0 && io->callback != NULL)
					io->callback(io->fd, cond, io->data);
			}

			cur = next;
		}
	}

	return (ret);
}
